use crate::func::context::*;

use super::*;

/// ```rust
/// # use radn::rcore::Context;
/// # use radn::rcore::Factory;
/// # use radn::rcore::FactoryExt;
/// # use radn::rcore::Mentionable;
/// # use radn::rstd::ResolverExt;
/// # use radn::rstd::SerializableExt;
/// # use radn::rstd::singular::SingularResolver;
/// # fn test<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>>(a: &A) {
/// let resolver = SingularResolver::from_mentionable(a).into_rc();
/// let bytes = a.bytes();
/// let factory = a.factory();
/// factory.parse_slice(&bytes, &resolver);
/// # }
/// ```
pub struct SingularResolver<'a, Ctx: Context<'a>> {
    points: Vec<Arc<dyn SingularResolution<'a, Ctx>>>,
}

struct SingularRobust<'a, Ctx: Context<'a>> {
    resolver: Arc<SingularResolver<'a, Ctx>>,
}

impl<'a, Ctx: Context<'a>> SingularResolver<'a, Ctx> {
    pub fn from_mentionable<A: Mentionable<'a, Ctx>>(a: &A) -> Self {
        let mut points = Vec::new();
        a.points_typed(&mut points);
        Self { points }
    }

    fn resolve_robust(
        self: Arc<Self>,
        address: Address,
    ) -> Result<HashResolution<'a, Ctx>, SingularError> {
        let point = self
            .points
            .get(address.index)
            .ok_or(SingularError::OutOfBounds {
                index: address.index,
                len: self.points.len(),
            })?;
        if point.s_hash() != address.point {
            Err(SingularError::Mismatch {
                index: address.index,
                expected: address.point,
                point: point.s_hash(),
            })
        } else {
            Ok(point.clone().singular())
        }
    }

    pub fn robust(self) -> Arc<dyn Resolver<'a, Ctx>>
    where
        Ctx: SingularCtx<'a>,
    {
        SingularRobust {
            resolver: self.into(),
        }
        .into_rc()
    }
}

impl<'a, Ctx: Context<'a>> Resolver<'a, Ctx> for SingularResolver<'a, Ctx> {
    fn resolve(self: Arc<Self>, address: Address) -> HashResolution<'a, Ctx> {
        self.resolve_robust(address).unwrap()
    }
}

pub trait SingularCtx<'a>: Context<'a> {
    fn from_singular(error: SingularError) -> Self::LookupError;
}

impl<'a, Ctx: Context<'a>> SingularCtx<'a> for Ctx
where
    Ctx::LookupError: From<SingularError>,
{
    fn from_singular(error: SingularError) -> Self::LookupError {
        error.into()
    }
}

impl<'a, Ctx: SingularCtx<'a>> Resolver<'a, Ctx> for SingularRobust<'a, Ctx> {
    fn resolve(self: Arc<Self>, address: Address) -> HashResolution<'a, Ctx> {
        self.resolver
            .clone()
            .resolve_robust(address)
            .map_err(Ctx::from_singular)
            .map_err(Err)
            .unwrap_or_else(Ctx::pure)
    }
}

impl<'a, Ctx: Context<'a>> PointsVisitor<'a, Ctx> for Vec<Arc<dyn SingularResolution<'a, Ctx>>> {
    fn visit<A: Mentionable<'a, Ctx>>(&mut self, point: &Point<'a, Ctx, A>) {
        self.push(Arc::new(point.clone()) as _);
    }
}
